#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#include <idsa_internal.h>

static char *_idsa_print_hex_digit = "0123456789ABCDEF";

/* ascii character manipulation */
#define isworkable(c)   (((c)>0x1f)&&((c)<0x80)&&((c)!='\\')&&((c)!='"')&&((c)!='^'))
#define ishigh(c)       ((c)&0x80)
#define printcontrol(c) ((c)+0x40)
#define printxdigit(c)  (_idsa_print_hex_digit[(c)&0x0f])

/****************************************************************************/

struct idsa_print_format {
  char *pf_name;
  IDSA_PRINT_FUNCTION pf_function;
};

static IDSA_PRINT_FORMAT _idsa_print_table[] = {
  {"native", idsa_print_native},
  {"internal", idsa_print_internal},
  {"syslog", idsa_print_syslog},
  {"ulm", idsa_print_ulm},
  {"tulm", idsa_print_tulm},
  {"csv", idsa_print_csv},
  {NULL, NULL}
};

IDSA_PRINT_FORMAT *idsa_print_lookup(char *n)
{

  int i = 0;
  while ((_idsa_print_table[i].pf_name != NULL)
	 && strcmp(_idsa_print_table[i].pf_name, n)) {
    i++;
  }

  if (_idsa_print_table[i].pf_name == NULL) {
    return NULL;
  } else {
    return &(_idsa_print_table[i]);
  }
}

int idsa_print(IDSA_EVENT * e, char *b, int l, IDSA_PRINT_FORMAT * p)
{
  if (p) {
    return (*(p->pf_function)) (e, b, l);
  } else {
    return -1;
  }
}

/****************************************************************************/

typedef int (*IDSA_PRINT_SINGLE) (IDSA_UNIT * u, char *s, int l);

static int idsa_print_time_colon(IDSA_UNIT * u, char *s, int l);
static int idsa_print_time_syslog(IDSA_UNIT * u, char *s, int l);
static int idsa_print_unit_value(IDSA_UNIT * u, char *s, int l);
static int idsa_print_unit_ulm(IDSA_UNIT * u, char *s, int l);
static int idsa_print_unit_tulm(IDSA_UNIT * u, char *s, int l);
/* could have a name=value print function */

#define IDSA_FORMAT_SINGLE  0x01	/* only process this one */
#define IDSA_FORMAT_TAIL    0x02	/* this one and all following ones */

struct idsa_print_item {
  int of_type;
  char *of_spacer;
  int of_number;
  IDSA_PRINT_SINGLE of_function;
};

static int idsa_print_unit(char *k, IDSA_UNIT * u, IDSA_PRINT_SINGLE f,
			   char *s, int l);
static int idsa_print_template(IDSA_EVENT * e, char *b, int l,
			       struct idsa_print_item *o);

static struct idsa_print_item idsa_format_ulm[] = {
  {IDSA_FORMAT_TAIL, " ", 0, &idsa_print_unit_ulm}
};

static struct idsa_print_item idsa_format_tulm[] = {
  {IDSA_FORMAT_TAIL, " ", 0, &idsa_print_unit_tulm}
};

static struct idsa_print_item idsa_format_csv[] = {
  {IDSA_FORMAT_TAIL, "\",\"", 0, &idsa_print_unit_value}
};

static struct idsa_print_item idsa_format_native[] = {
  {IDSA_FORMAT_SINGLE, " ", IDSA_N_TIME, &idsa_print_time_colon},
  {IDSA_FORMAT_SINGLE, " ", IDSA_N_HOST, &idsa_print_unit_value},
  {IDSA_FORMAT_SINGLE, " ", IDSA_N_UID, &idsa_print_unit_value},
  {IDSA_FORMAT_SINGLE, ":", IDSA_N_GID, &idsa_print_unit_value},
  {IDSA_FORMAT_SINGLE, " ", IDSA_N_SERVICE, &idsa_print_unit_value},
  {IDSA_FORMAT_SINGLE, "[", IDSA_N_PID, &idsa_print_unit_value},
  {IDSA_FORMAT_SINGLE, "] ", IDSA_N_SCHEME, &idsa_print_unit_value},
  {IDSA_FORMAT_SINGLE, ".", IDSA_N_NAME, &idsa_print_unit_value},
  {IDSA_FORMAT_SINGLE, " ", IDSA_N_ARISK, &idsa_print_unit_value},
  {IDSA_FORMAT_SINGLE, ":", IDSA_N_CRISK, &idsa_print_unit_value},
  {IDSA_FORMAT_SINGLE, ":", IDSA_N_IRISK, &idsa_print_unit_value},
  {IDSA_FORMAT_TAIL, " ", IDSA_M_REQUIRED, &idsa_print_unit_ulm}
};

static struct idsa_print_item idsa_format_syslog[] = {
  {IDSA_FORMAT_SINGLE, NULL, IDSA_N_TIME, &idsa_print_time_syslog},
  {IDSA_FORMAT_SINGLE, " ", IDSA_N_UID, &idsa_print_unit_value},
  {IDSA_FORMAT_SINGLE, "@", IDSA_N_HOST, &idsa_print_unit_value},
  {IDSA_FORMAT_SINGLE, " ", IDSA_N_SERVICE, &idsa_print_unit_value},
  {IDSA_FORMAT_SINGLE, "[", IDSA_N_PID, &idsa_print_unit_value},
  {IDSA_FORMAT_SINGLE, "]: ", IDSA_N_NAME, &idsa_print_unit_value},
  {IDSA_FORMAT_TAIL, " ", IDSA_M_REQUIRED, &idsa_print_unit_ulm}
};

/****************************************************************************/

static int idsa_print_time_syslog(IDSA_UNIT * u, char *s, int l)
{
  time_t t;
  static time_t savet = 0;
  static char saves[16] = "Jan  1 00:00:00";
  struct tm *tp;
  const char *months[] =
      { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep",
    "Oct", "Nov", "Dec"
  };

#ifdef TRACE
  fprintf(stderr, "idsa_print_time_syslog(): previous time is <%s>\n",
	  saves);
#endif

  if (l < 16) {
    return -1;
  } else {

    idsa_unit_get(u, &t, sizeof(time_t));

    if (t != savet) {
      tp = localtime(&t);
      snprintf(saves, 16, "%s %2d %02d:%02d:%02d", months[tp->tm_mon % 12],
	       tp->tm_mday, tp->tm_hour, tp->tm_min, tp->tm_sec);
      savet = t;
    }
#ifdef TRACE
    fprintf(stderr, "idsa_print_time_syslog(): new time is <%s>\n", saves);
#endif
    memcpy(s, saves, 15);
    return 15;
  }
}

static int idsa_print_time_colon(IDSA_UNIT * u, char *s, int l)
{
  time_t t;
  static time_t savet = 0;
  static char saves[20] = "1970:01:01:00:00:00";
  struct tm *tp;

#ifdef TRACE
  fprintf(stderr, "idsa_print_time_colon(): previous time is <%s>\n",
	  saves);
#endif

  if (l < 21) {
    return -1;
  } else {

    idsa_unit_get(u, &t, sizeof(time_t));

    if (t != savet) {
      tp = localtime(&t);
      snprintf(saves, 20, "%04d:%02d:%02d:%02d:%02d:%02d",
	       tp->tm_year + 1900, tp->tm_mon + 1, tp->tm_mday,
	       tp->tm_hour, tp->tm_min, tp->tm_sec);
      savet = t;
    }
#ifdef TRACE
    fprintf(stderr, "idsa_print_time_colon(): new time is <%s>\n", saves);
#endif
    memcpy(s, saves, 19);
    return 19;
  }
}

static int idsa_print_unit_value(IDSA_UNIT * u, char *s, int l)
{
  int result, i;
  unsigned char *p;

  /* yuck, cast to unsigned to stay in range */
  p = s;

  result = idsa_unit_print(u, p, l, 0);

  /* might also want a test to see which types need not be escaped */

  if (result > 0) {
    for (i = 0; i < result; i++) {
      if (!isworkable(p[i])) {
	if (ishigh(p[i])) {	/* high characters are escaped as \xx */
	  if (result + 1 < l) {
	    memmove(p + i + 2, p + i, result - i);
	    p[i + 1] = printxdigit((p[i] & 0xf0) >> 4);
	    p[i + 2] = printxdigit(p[i] & 0x0f);
	    p[i] = '\\';
	    i += 2;
	    result += 2;
	  } else {
	    result = (-1);
	  }
	} else {		/* ^c or \c */
	  if (result < l) {
	    memmove(p + i + 1, p + i, result - i);
	    if (iscntrl(p[i])) {	/* control characters escaped as ^c */
	      p[i + 1] = printcontrol(p[i]);
	      p[i] = '^';
	    } else {		/* escape escape characters */
	      p[i] = '\\';
	    }
	    i++;
	    result++;
	  } else {		/* buffer too short */
	    result = (-1);
	  }
	}
      }				/* else ok character, no escape needed */
    }

#ifdef TRACE
    fprintf(stderr, "idsa_print_unit_value(): escaped <");
    for (i = 0; i < result; i++)
      fputc(s[i], stderr);
    fprintf(stderr, ">\n");
#endif

  }
  return result;
}

static int idsa_print_unit_ulm(IDSA_UNIT * u, char *s, int l)
{
  int result = 0;
  int k;
  char *n;

  if (u == NULL) {
    return -1;
  }

  n = idsa_unit_name(u);
  for (k = 0; (result < l) && (n[k] != '\0'); k++, result++) {
    s[result] = n[k];
  }

  if ((result + 3) >= l) {
    return -1;
  }
  s[result++] = '=';
  s[result++] = '"';

  k = idsa_print_unit_value(u, s + result, l - result);
  if (k < 0) {
    return -1;
  }
  result += k;

  if (result >= l) {
    return -1;
  }
  s[result++] = '"';

  return result;
}

static int idsa_print_unit_tulm(IDSA_UNIT * u, char *s, int l)
{
  int result = 0;
  int k;
  char *t, *n;

  if (u == NULL) {
    return -1;
  }

  n = idsa_unit_name(u);
  for (k = 0; (result < l) && (n[k] != '\0'); k++, result++) {
    s[result] = n[k];
  }

  t = idsa_type_name(idsa_unit_type(u));
  k = strlen(t);
  if ((result + k + 4) >= l) {
    return -1;
  }
  s[result++] = ':';
  strncpy(s + result, t, k);
  result += k;
  s[result++] = '=';
  s[result++] = '"';

  k = idsa_print_unit_value(u, s + result, l - result);
  if (k < 0) {
    return -1;
  }
  result += k;

  if (result >= l) {
    return -1;
  }
  s[result++] = '"';

  return result;
}

/****************************************************************************/

static int idsa_print_unit(char *k, IDSA_UNIT * u, IDSA_PRINT_SINGLE f,
			   char *s, int l)
{
  int i, r;

  if (k) {
    for (i = 0; (i < l) && (k[i] != '\0'); i++) {
      s[i] = k[i];
    }
    if (k[i] != '\0') {
      return -1;
    }
  } else {
    i = 0;
  }

  if (u) {
    r = (*(f)) (u, s + i, l - i);
    if (r < 0) {
      return -1;
    }
  } else {
    r = 0;
  }

  return i + r;
}

static int idsa_print_template(IDSA_EVENT * e, char *b, int l,
			       struct idsa_print_item *o)
{
  int result = 0;
  int i, j, k;

  char *tbuf;
  int tlen;

  tlen = l;
  tbuf = b;

  i = 0;
  while (o[i].of_type == IDSA_FORMAT_SINGLE) {

    j = idsa_print_unit(o[i].of_spacer,
			idsa_event_unitbynumber(e, o[i].of_number),
			o[i].of_function, tbuf, tlen);

    if (j < 0) {
      return -1;
    }
    result += j;
    tbuf = tbuf + j;
    tlen = tlen - j;

    i++;
  }

#ifdef TRACE
  fprintf(stderr,
	  "idsa_print_template(): finished singles, now processing tail\n");
#endif

  if (o[i].of_type == IDSA_FORMAT_TAIL) {
    for (k = o[i].of_number; k < e->e_count; k++) {
      j =
	  idsa_print_unit(o[i].of_spacer, idsa_event_unitbynumber(e, k),
			  o[i].of_function, tbuf, tlen);

      if (j < 0) {
	return -1;
      }
      result += j;
      tbuf = tbuf + j;
      tlen = tlen - j;
    }
  }
  return result;
}

/****************************************************************************/

int idsa_print_internal(IDSA_EVENT * e, char *b, int l)
{
  return idsa_event_tobuffer(e, b, l);
}

int idsa_print_ulm(IDSA_EVENT * e, char *b, int l)
{
  int result = 0;
  int i;

  i =
      snprintf(b, l, "allow=\"%s\" repeat=\"%d\"",
	       (e->e_result & IDSA_L_DENY) ? "false" : "true",
	       e->e_repeat);
  if (i < 0) {
    return -1;
  }
  result += i;

  i = idsa_print_template(e, b + result, l - result, idsa_format_ulm);
  if (i < 0) {
    return -1;
  }
  result += i;

  if (result >= l) {
    return -1;
  }
  b[result++] = '\n';

  return result;
}

int idsa_print_tulm(IDSA_EVENT * e, char *b, int l)
{
  int result = 0;
  int i;

  i =
      snprintf(b, l, "allow:flag=\"%s\" repeat:integer=\"%d\"",
	       (e->e_result & IDSA_L_DENY) ? "false" : "true",
	       e->e_repeat);
  if (i < 0) {
    return -1;
  }
  result += i;

  i = idsa_print_template(e, b + result, l - result, idsa_format_tulm);
  if (i < 0) {
    return -1;
  }
  result += i;

  if (result >= l) {
    return -1;
  }
  b[result++] = '\n';

  return result;
}

int idsa_print_csv(IDSA_EVENT * e, char *b, int l)
{
  int result = 0;
  int i;

  i =
      snprintf(b, l, "\"%s\",\"%d",
	       (e->e_result & IDSA_L_DENY) ? "false" : "true",
	       e->e_repeat);
  if (i < 0) {
    return -1;
  }
  result += i;

  i = idsa_print_template(e, b + result, l - result, idsa_format_csv);
  if (i < 0) {
    return -1;
  }
  result += i;

  if (result >= l) {
    return -1;
  }
  b[result++] = '"';

  if (result >= l) {
    return -1;
  }
  b[result++] = '\n';

  return result;
}

int idsa_print_native(IDSA_EVENT * e, char *b, int l)
{
  int result = 0;
  int i;

  i =
      snprintf(b, l, "%c%d", (e->e_result & IDSA_L_DENY) ? '-' : '+',
	       e->e_repeat);
  if (i < 0) {
    return -1;
  }
  result += i;

  i = idsa_print_template(e, b + result, l - result, idsa_format_native);
  if (i < 0) {
    return -1;
  }
  result += i;

  if (result >= l) {
    return -1;
  }
  b[result++] = '\n';

  return result;
}

int idsa_print_syslog(IDSA_EVENT * e, char *b, int l)
{
  int result = 0;
  int i;

  i = idsa_print_template(e, b + result, l - result, idsa_format_syslog);
  if (i < 0) {
    return -1;
  }
  result += i;

  if (result >= l) {
    return -1;
  }
  b[result++] = '\n';

  return result;
}
