#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sched.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>

#include <sys/time.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "misc.h"

#include "mod.h"
#include "rule.h"
#include "branch.h"
#include "set.h"
#include "parse.h"

static void mod_pipe(STATE_SET * s, MOD_PIPE * md);
static MOD_PIPE *mod_new(STATE_SET * s, char *command, int sec, int usec,
			 int failmode);

static void mod_pipe(STATE_SET * s, MOD_PIPE * md)
{
  int pp[2];
  int status;
  int flags;
  int efd;

  if (socketpair(AF_UNIX, SOCK_STREAM, 0, pp)) {
    parse_error(s, "error: unable to create pipe for %s: %s\n",
		md->m_command, strerror(errno));
    return;
  }
  md->m_pid = fork();
  switch (md->m_pid) {
  case -1:
    parse_error(s, "error: unable to fork %s: %s\n", md->m_command,
		strerror(errno));
    close(pp[0]);
    close(pp[1]);
    break;
  case 0:			/* in child */

    close(pp[0]);

    if (pp[1] != STDIN_FILENO) {
      if (dup2(pp[1], STDIN_FILENO) != STDIN_FILENO) {	/* failure */
	/* WARNING: abuse of return code */
	exit(errno);
      }
    }
    if (pp[1] != STDOUT_FILENO) {
      if (dup2(pp[1], STDOUT_FILENO) != STDOUT_FILENO) {	/* failure */
	/* WARNING: abuse of return code */
	exit(errno);
      }
    }
    if (pp[1] >= STDERR_FILENO) {
      close(pp[1]);
    }

    close(STDERR_FILENO);
    efd = open("/dev/null", O_WRONLY);
    if ((efd > 0) && (efd != STDERR_FILENO)) {
      dup2(efd, STDERR_FILENO);
      close(efd);
    }

    strexec(md->m_command);
    /* WARNING: abuse of return code */
    exit(errno);

    break;
  default:			/* in parent */
    close(pp[1]);

    md->m_fd = pp[0];

    fcntl(md->m_fd, F_SETFD, FD_CLOEXEC);

    flags = fcntl(md->m_fd, F_GETFL, 0);

    if ((flags == (-1)) |
	(fcntl(md->m_fd, F_SETFL, O_NONBLOCK | flags) == (-1))) {
      parse_error(s, "error: unable to make pipe to %s nonblocking: %s\n",
		  md->m_command, strerror(errno));
      close(md->m_fd);
    }
    /* yield to child */
    sched_yield();

    if (waitpid(md->m_pid, &status, WNOHANG)) {	/* found a child or error */
      if (WIFEXITED(status)) {
	parse_error(s, "error: unable to start child process %s: %s\n",
		    md->m_command, strerror(WEXITSTATUS(status)));
      } else if (WIFSIGNALED(status)) {
	parse_error(s, "error: child process %s killed by signal %d\n",
		    md->m_command, WTERMSIG(status));
      }
      close(md->m_fd);
    }
    break;
  }

}

static MOD_PIPE *mod_new(STATE_SET * s, char *command, int sec, int usec,
			 int failmode)
{
  MOD_PIPE *md;

  md = malloc(sizeof(MOD_PIPE));
  if (md) {
    md->m_failmode = failmode;
    md->m_times = sec;
    md->m_timeu = usec;

    strncpy(md->m_command, command, IDSA_M_MEDIUM - 1);
    md->m_command[IDSA_M_MEDIUM - 1] = '\0';

#ifdef TRACE
    fprintf(stderr, "mod_new(): timemout is %ds + %us\n", md->m_times,
	    md->m_timeu);
#endif

    md->m_interrupt = 0;
    md->m_count = 0;
    md->m_match = NULL;

    md->m_event = idsa_event_new(0);
    if (md->m_event) {
      /* if mod_pipe fails, return error via s->s_err* */
      mod_pipe(s, md);
    } else {
      free(md);
      md = NULL;
    }
  }
  return md;
}

void mod_free(MOD_PIPE * md)
{
  if (md->m_match) {
    free(md->m_match);
    md->m_match = NULL;
  }
  md->m_count = 0;

  if (md->m_event) {
    idsa_event_free(md->m_event);
    md->m_event = NULL;
  }
  close(md->m_fd);
  md->m_fd = (-1);

  /* a kill(md->m_pid) would be excessive */

  free(md);
}

int mod_trigger(MOD_PIPE * md, IDSA_EVENT * e)
{
  char buffer[IDSA_M_MESSAGE];
  int should_write, write_result, have_written, read_result, have_read,
      copy_bytes, still_running;

  int result = 0;
  int code;
  int i;

  fd_set fs;
  struct timeval tv;

  if (md->m_failmode == MOD_FAILOPEN) {
    code = IDSA_L_ALLOW;
  } else {
    code = IDSA_L_DENY;
  }

#ifdef TRACE
  fprintf(stderr, "mod_trigger(): triggering\n");
#endif

  tv.tv_sec = md->m_times;
  tv.tv_usec = md->m_timeu;

  /* attempt to recover from previous stuffups */

  if (md->m_interrupt) {
    have_read = 0;
    still_running = 1;
    while (still_running) {
      read_result =
	  read(md->m_fd, buffer + have_read, IDSA_M_MESSAGE - have_read);
      switch (read_result) {
      case -1:
	switch (errno) {
	case EAGAIN:		/* wait, but only briefly */
	  FD_ZERO(&fs);
	  FD_SET(md->m_fd, &fs);
	  if (select(md->m_fd + 1, &fs, NULL, NULL, &tv) != 1) {
	    still_running = 0;
	  }
	  break;
	case EINTR:		/* interrupt, try again */
	  break;
	default:		/* all others pose problems */
	  still_running = 0;
	  break;
	}
	break;
      case 0:			/* ouch, eof */
	still_running = 0;
	break;
      default:			/* something to read */
	still_running = 0;
	have_read += read_result;
	do {
	  copy_bytes = idsa_event_frombuffer(e, buffer, have_read);
	  if (copy_bytes > 0) {
	    still_running = 1;
	    if (md->m_interrupt > 0) {
	      md->m_interrupt = md->m_interrupt - 1;
	    }
	    if (copy_bytes < have_read) {
	      memmove(buffer, buffer + copy_bytes, have_read - copy_bytes);
	      have_read -= copy_bytes;
	    } else {
	      have_read = 0;
	    }
	  }
	} while (copy_bytes > 0);
	break;
      }
    }				/* end of loop */

    if (have_read > 0) {	/* some gunk still left, assume corruption */
      while (read(md->m_fd, buffer, IDSA_M_MESSAGE) > 0);
      md->m_interrupt = 0;
    }				/* nothing in buffer, probably a good thing */
  }
  /* now hopefully pipe clean for our own write/read */

  if (md->m_interrupt == 0) {	/* managed to recover or no error at all */
    should_write = idsa_event_tobuffer(e, buffer, IDSA_M_MESSAGE);
    if (should_write > 0) {	/* something to write */

      write_result = write(md->m_fd, buffer, should_write);
      if (write_result == should_write) {
	have_written = write_result;
      } else if ((write_result < 0) && (errno != EAGAIN)
		 && (errno != EINTR)) {
	have_written = 0;
      } else {
	if (write_result < 0) {
	  have_written = 0;
	} else {
	  have_written = write_result;
	}
	FD_ZERO(&fs);
	FD_SET(md->m_fd, &fs);
	if (select(md->m_fd + 1, NULL, &fs, NULL, &tv) != 0) {
	  do {
	    write_result =
		write(md->m_fd, buffer + have_written,
		      should_write - have_written);
	  } while ((write_result < 0) && (errno == EINTR));
	  if (write_result > 0) {
	    have_written += write_result;
	  }
	}
      }

      if (have_written == should_write) {	/* write ok, something to read */
	read_result = read(md->m_fd, buffer, IDSA_M_MESSAGE);
	if ((read_result < 0) && ((errno == EAGAIN) || (errno == EINTR))) {	/* try harder */
	  FD_ZERO(&fs);
	  FD_SET(md->m_fd, &fs);
	  if (select(md->m_fd + 1, &fs, NULL, NULL, &tv) != 0) {
	    read_result = read(md->m_fd, buffer, IDSA_M_MESSAGE);
	  }
	}
	if (read_result > 0) {	/* something to read */
	  copy_bytes =
	      idsa_event_frombuffer(md->m_event, buffer, read_result);
	  if (copy_bytes == read_result) {	/* exactly one event, not more, not less */
	    if (md->m_event->e_result & IDSA_L_DENY) {
	      code = IDSA_L_DENY;
	    } else {
	      code = IDSA_L_ALLOW;
	    }
	  }			/* else no valid event */
	} else {		/* nothing to read, tell child to hurry up */
	  md->m_interrupt = 1;
	  kill(md->m_pid, SIGINT);
	}
      }				/* else write failed */
    }				/* else copy to buffer failed */
  }				/* else unable to clean pipe */
  if (code == IDSA_L_ALLOW) {
    for (i = 0; i < md->m_count; i++) {
      match_fire_true(md->m_match[i]);
    }
  } else {
    for (i = 0; i < md->m_count; i++) {
      match_fire_false(md->m_match[i]);
    }
  }

  return result;
}

RULE_MATCH *mod_match_make(STATE_SET * s, char *command, int sec, int usec,
			   int failmode)
{
  UNION_BRANCH *b;
  MOD_PIPE *md;
  RULE_MATCH *m, **q;

  m = NULL;
  md = NULL;
  b = s->s_branch;
  while (b) {
    if (b->b_type == BRANCH_TYPE_TABLE) {
      md = b->b_un.b_mod;
      if (strncmp(md->m_command, command, IDSA_M_MEDIUM)) {
	b = b->b_next;
	md = NULL;
      } else {
#ifdef TRACE
	fprintf(stderr, "mod_match_make(): match\n");
#endif
	b->b_last = s->s_currentrule;
	b = NULL;
      }
    } else {
      b = b->b_next;
    }
  }

  if (md == NULL) {		/* not found, create */
    md = mod_new(s, command, sec, usec, failmode);
    if (md) {
      b = branch_new(s->s_currentrule, BRANCH_TYPE_MOD);
      if (b) {
	b->b_un.b_mod = md;
	branch_append(s, b);
      } else {
	parse_errormem(s, sizeof(UNION_BRANCH));
	mod_free(md);
	md = NULL;
      }
    } else {
      parse_errormem(s, sizeof(MOD_PIPE));
    }
  }
  if (md) {			/* add match to this entry */
    m = match_new(RULE_GATE_NONE);
    if (m) {
#ifdef TRACE
      fprintf(stderr,
	      "mod_match_make(): about to resize %p from %d to %d\n",
	      md->m_match, md->m_count, md->m_count + 1);
#endif
      q = realloc(md->m_match, sizeof(RULE_MATCH *) * (md->m_count + 1));
      if (q) {
	q[md->m_count] = m;
	md->m_count++;
	md->m_match = q;
      } else {
	parse_errormem(s, sizeof(RULE_MATCH *) * (md->m_count + 1));
	match_free(m);
	m = NULL;
      }
    } else {
      parse_errormem(s, sizeof(RULE_MATCH));
    }
  }
  return m;
}
